/*
 * Copyright (c) 2016-2017 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
#include <stdio.h>
#include <unistd.h>
#include <stdlib.h>
#include <pthread.h>
#include <sys/types.h>
#include <sys/inotify.h>
#include <fcntl.h>
#include <json.h>
#include <errno.h>
#include "sa_common.h"
#include "sa_types.h"
#include "parse_file.h"

static int json_get_int_from_obj(json_object * inputObj, char *key)
{
	struct json_object *bodyObj;
	enum json_type type;
	int ret = 0;

	type = json_object_get_type(inputObj);

	if (type == json_type_object) {
		if (json_object_object_get_ex(inputObj, key, &bodyObj))
			ret = json_object_get_int(bodyObj);
	}
	return ret;
}

static int json_get_boolean_from_obj(json_object * inputObj, char *key)
{
	struct json_object *bodyObj;
	enum json_type type;
	int ret = 0;

	type = json_object_get_type(inputObj);

	if (type == json_type_object) {
		if (json_object_object_get_ex(inputObj, key, &bodyObj))
			ret = json_object_get_boolean(bodyObj);
	}
	return ret;
}

static char *json_get_string_from_obj(json_object * inputObj, char *key)
{
	const char *buf = NULL;
	char *ret_buf = NULL;
	json_object *elementObj = NULL;
	int i = 0, j = 0;

	if (json_object_object_get_ex(inputObj, key, &elementObj)) {
		buf = json_object_to_json_string_ext(elementObj, JSON_C_TO_STRING_SPACED | JSON_C_TO_STRING_PRETTY);
		ret_buf = malloc(strlen(buf) - 1);
		if (ret_buf != NULL) {
			memset(ret_buf, 0x00, strlen(buf) - 1);	// "<-- exclude
			for (i = 1, j = 0; i < strlen(buf) - 1; i++) {	// "\ <-- exclude
				if (buf[i] == '\\')
					continue;
				ret_buf[j++] = buf[i];
			}
		} else {
			return NULL;
		}
	} else {
		return NULL;
	}

	_D("string object [%s]", ret_buf);
	return ret_buf;
}

static void print_system_config(sa_system_s * systemData)
{
	int i;

	_D("sa_system_s[%x]", systemData);
	if (systemData != NULL) {
		if (systemData->proxy != NULL) {
			_D("systemData::httpProxy [%s]", systemData->proxy->httpProxy);
			_D("systemData::httpsProxy [%s]", systemData->proxy->httpsProxy);
		}

		if (systemData->dockerdOpt != NULL) {
			if (systemData->dockerdOpt->count > 0) {
				for (i = 0; i < systemData->dockerdOpt->count; i++) {
					if (systemData->dockerdOpt->insecureRegistries[i] != NULL)
						_D("systemData::insecureRegistries[%d] [%s]", i, systemData->dockerdOpt->insecureRegistries[i]);
				}
			}
		}
	}

}

static void print_config_info(sa_config_s * config)
{
	if (config != NULL) {
		_D("Version [%s]", config->version);

		if (config->systemData)
			print_system_config(config->systemData);
	}
}

static int __parse_system_insecure_registries(json_object * inputObj, sa_dockerd_opt_s * opt)
{
	int ret = 0;
	char *value = NULL;
	struct json_object *registryObj = NULL;
	int ulIndex = 0;
	int str_len = 0;

	if (inputObj == NULL || opt == NULL) {
		_D("__parse_system_insecure_registries input error");
		return -1;
	}

	memset(opt, 0x00, sizeof(sa_dockerd_opt_s));

	opt->count = json_object_array_length(inputObj);
	if (opt->count > 0) {
		opt->insecureRegistries = (char **)malloc(sizeof(char *) * opt->count);
		if (opt->insecureRegistries != NULL) {
			for (ulIndex = 0; ulIndex < opt->count; ulIndex++) {
				registryObj = json_object_array_get_idx(inputObj, ulIndex);
				if (registryObj != NULL) {
					value = json_object_get_string(registryObj);
					if (value != NULL) {
						str_len = strlen(value);
						_D("value[%d]: %s (%d)", ulIndex, value, str_len);
						opt->insecureRegistries[ulIndex] = (char *)malloc(str_len + 1);
						if (opt->insecureRegistries[ulIndex] != NULL) {
							snprintf(opt->insecureRegistries[ulIndex], str_len + 1, "%s", value);
							_D("insecure [%d]: %s", ulIndex, opt->insecureRegistries[ulIndex]);
						} else {
							ret = -1;
							_E("malloc failed for insecureRegistries second");
							break;
						}
					}
				}
			}
		} else {
			ret = -1;
			_E("malloc failed for inscureRegistries first");
		}
	} else {
		_D("Insecure registries is empty");
	}

	return ret;
}

static int __parse_system_data(json_object * inputObj, sa_config_s * setupConfig)
{
	int ret = 0;
	char *httpProxy = NULL;
	char *httpsProxy = NULL;
	struct json_object *arrayObj = NULL;

	if (inputObj == NULL) {
		_D("__parse_system_data input error");
		return -1;
	}

	setupConfig->systemData = (sa_system_s *) malloc(sizeof(sa_system_s));

	if (setupConfig->systemData != NULL) {
		memset(setupConfig->systemData, 0x00, sizeof(sa_system_s));

		httpProxy = json_get_string_from_obj(inputObj, JSON_STR_ID_HTTP_PROXY);
		httpsProxy = json_get_string_from_obj(inputObj, JSON_STR_ID_HTTPS_PROXY);
		if ((httpProxy != NULL) || (httpsProxy != NULL))
			setupConfig->systemData->proxy = (sa_proxy_s *) malloc(sizeof(sa_proxy_s));

		//httpProxy
		if (httpProxy != NULL) {
			if (strlen(httpProxy) > 0) {
				if (setupConfig->systemData->proxy != NULL) {
					snprintf(setupConfig->systemData->proxy->httpProxy, sizeof(setupConfig->systemData->proxy->httpProxy), "%s", httpProxy);
				} else {
					ret = -1;
					_D("malloc fail setupConfig->systemData->proxy");
				}
			}
		}
		//httpsProxy
		if (httpsProxy != NULL) {
			if (strlen(httpsProxy) > 0) {
				if (setupConfig->systemData->proxy != NULL) {
					snprintf(setupConfig->systemData->proxy->httpsProxy, sizeof(setupConfig->systemData->proxy->httpsProxy), "%s", httpsProxy);
				} else {
					ret = -1;
					_D("malloc fail setupConfig->systemData->proxy");
				}
			}
		}
		// insecure-registries
		if (json_object_object_get_ex(inputObj, JSON_STR_ID_INSECURE_REGISTRIES, &arrayObj)) {
			if (json_object_get_type(arrayObj) == json_type_array) {
				setupConfig->systemData->dockerdOpt = (sa_dockerd_opt_s *) malloc(sizeof(sa_dockerd_opt_s));
				if (setupConfig->systemData->dockerdOpt != NULL) {
					ret = __parse_system_insecure_registries(arrayObj, setupConfig->systemData->dockerdOpt);
				} else {
					_E("malloc failed setupConfig->systemData->dockerdOpt");
					ret = -1;
				}
			} else {
				_E("The type of insecure registries is wrong");
				ret = -1;
			}
		}

		if (httpProxy != NULL) {
			free(httpProxy);
			httpProxy = NULL;
		}
		if (httpsProxy != NULL) {
			free(httpsProxy);
			httpsProxy = NULL;
		}
	} else {
		ret = -1;
		_E("malloc failed for system data");
	}

	return ret;
}

static int __parse_version(json_object * inputObj, sa_config_s * setupConfig)
{
	char *version = NULL;

	if (inputObj == NULL) {
		_D("__parse_version input error");
		return -1;
	}

	version = json_get_string_from_obj(inputObj, JSON_STR_ID_VERSION);

	if (version != NULL) {
		memcpy(setupConfig->version, version, MIN(strlen(version), sizeof(setupConfig->version) - 1));
		free(version);
		version = NULL;
	}

	return 0;
}

static int __parse_config(char *file, sa_config_s * setupConfig)
{
	struct json_object *configObj = NULL;
	char *jsonData = NULL;
	int readLen = 0;
	int fd = 0;
	int len = 0;
	int ret = 0;

	fd = open(file, O_RDONLY);
	if (fd < 0) {
		_E("config file can't be opened");
		return -1;
	}

	len = lseek(fd, 0L, SEEK_END);

	if (len > 0) {
		if (lseek(fd, 0L, SEEK_SET) < 0) {
			ret = -1;
			_E("find beginning of file returns error");
		} else {
			jsonData = (char *)malloc(len + 1);
			if (jsonData != NULL) {
				memset(jsonData, 0x00, len + 1);
				readLen = read(fd, jsonData, len);
				_D("JSON full data[%s]", jsonData);

				configObj = json_tokener_parse(jsonData);

				if (configObj != NULL && readLen > 0) {
					// parse version
					ret = __parse_version(configObj, setupConfig);
					if (ret == 0)
						ret = __parse_system_data(configObj, setupConfig);
				} else {
					ret = -1;
					_D("ConfigObj is not existed");
				}

				json_object_put(configObj);
				free(jsonData);
			} else {
				ret = -1;
				_D("memory allocation fail for jsonData");
			}
		}
	}

	close(fd);

	return ret;
}

sa_error_e sa_inputfile_get_config_info(sa_config_s * config, char *file_name)
{
	sa_error_e ret = SA_ERROR_NONE;
	// check file
	if (access(file_name, 0) != 0)
		return SA_ERROR_NOT_SUPPORTED;

	// parsing and fill into struct
	ret = __parse_config(file_name, config);
	if (ret != 0) {
		_D("Config parsing error");
		return SA_ERROR_PARSING;
	}

	print_config_info(config);

	return ret;
}

static void release_system_resource(sa_system_s * system)
{
	int ulIndex = 0;
	if (system != NULL) {
		if (system->proxy != NULL)
			free(system->proxy);

		if (system->dockerdOpt != NULL) {
			if (system->dockerdOpt->insecureRegistries != NULL) {
				if (system->dockerdOpt->count > 0) {
					for (ulIndex = 0; ulIndex < system->dockerdOpt->count; ulIndex++) {
						if (system->dockerdOpt->insecureRegistries[ulIndex] != NULL)
							free(system->dockerdOpt->insecureRegistries[ulIndex]);
					}
				}
				free(system->dockerdOpt->insecureRegistries);
			}
			free(system->dockerdOpt);
		}
		free(system);
	}
}

void sa_inputfile_release_resource(sa_config_s * config)
{

	if (config != NULL) {
		if (config->systemData != NULL)
			release_system_resource(config->systemData);
	}
}

int __get_jsonobj_from_file(struct json_object **jsonObj, char *file_name)
{
	char *jsonData = NULL;
	int readLen = 0, fd = 0, len = 0, ret = 0;

	fd = open(file_name, O_RDONLY);
	if (fd < 0) {
		_E("config file can't be opened");
		return -1;
	}

	len = lseek(fd, 0L, SEEK_END);

	if (len > 0) {
		if (lseek(fd, 0L, SEEK_SET) < 0) {
			ret = -1;
			_E("find beginning of file returns error");
		} else {
			jsonData = (char *)malloc(len + 1);
			if (jsonData != NULL) {
				memset(jsonData, 0x00, len + 1);
				readLen = read(fd, jsonData, len);
				*jsonObj = json_tokener_parse(jsonData);

				if (*jsonObj == NULL || readLen <= 0) {
					ret = -1;
					_D("ConfigObj is not existed");
				}
				free(jsonData);
			} else {
				ret = -1;
				_D("memory allocation fail for jsonData");
			}
		}
	}
	close(fd);
	return ret;
}

int __create_system_config_file(struct json_object *jsonObj, char *write_file)
{
	int ret = 0;

	if (jsonObj) {
		struct json_object *jsonObjVer = NULL;	// extract version object

		// prepare json objects
		json_object_object_get_ex(jsonObj, "version", &jsonObjVer);
		json_object_object_del(jsonObj, "wifi");
		json_object_object_del(jsonObj, "ethernet");

		json_object_to_file_ext(write_file, jsonObj, JSON_C_TO_STRING_SPACED | JSON_C_TO_STRING_PRETTY);
	} else {
		ret = -1;
	}

	return ret;
}

int __create_target_config_file(struct json_object *jsonObj, char *write_file)
{
	int ret = 0;

	if (jsonObj) {
		struct json_object *jsonObjVer = NULL;	// extract version object
		struct json_object *jsonObjName = NULL;	// extract deviceName object

		// prepare json objects
		json_object_object_get_ex(jsonObj, "version", &jsonObjVer);
		json_object_object_get_ex(jsonObj, "deviceName", &jsonObjName);

		// result string for target
		struct json_object *jsonObjResultTarget = json_object_new_object();
		json_object_object_add(jsonObjResultTarget, "version", jsonObjVer);
		if (jsonObjName)
			json_object_object_add(jsonObjResultTarget, "deviceName", jsonObjName);

		json_object_to_file_ext(write_file, jsonObjResultTarget, JSON_C_TO_STRING_SPACED | JSON_C_TO_STRING_PRETTY);
	} else {
		ret = -1;
	}

	return ret;
}

void sa_split_files(char *sourceFile, char *dstFile_system, char *dstFile_target)
{
	struct json_object *jsonObj = NULL;
	char *baseString = "{ \"version\": \"0.1\", \
							\"system\": { }, \
							\"deviceName\": { }, \
						}";
	if (0 != __get_jsonobj_from_file(&jsonObj, sourceFile) || NULL == jsonObj) {
		// if there is no base json object, create with skeleton string.
		jsonObj = json_tokener_parse(baseString);
		_D("base json obj \n%s\n", json_object_to_json_string_ext(jsonObj, JSON_C_TO_STRING_SPACED | JSON_C_TO_STRING_PRETTY));
	}

	if (jsonObj) {
		if (dstFile_target)
			__create_target_config_file(jsonObj, dstFile_target);

		// system should be located in last as it has remove operation in json object
		if (dstFile_system)
			__create_system_config_file(jsonObj, dstFile_system);

		json_object_put(jsonObj);
	}
}
